@@ -1,5 +1,7 @@
 #include "pki.h"
 #include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
 
 #define USE_RINTERNALS 1
 #include <Rinternals.h>
@@ -15,7 +17,7 @@
 	 as an option. */
 
 /* from init.c */
-void PKI_init();
+void PKI_init(void);
 
 /* OpenSSL 1.1 has changed APIs - adapt accordingly */
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
@@ -49,11 +51,11 @@ SEXP PKI_load_DER_X509(SEXP what) {
 	Rf_error("%s", ERR_error_string(ERR_get_error(), NULL));
     res = PROTECT(R_MakeExternalPtr(x509, R_NilValue, R_NilValue));
     R_RegisterCFinalizerEx(res, PKI_free_X509, TRUE);
-    setAttrib(res, install("class"), mkString("X509cert"));
+    setAttrib(res, R_ClassSymbol, PROTECT(mkString("X509cert")));
     /* we add the content to the cert in case someone tries to serialize it */
-    ia = install("crt.DER");
+    ia = PROTECT(install("crt.DER"));
     setAttrib(res, ia, what);
-    UNPROTECT(1);
+    UNPROTECT(3);
     return res;
 }
 
@@ -82,7 +84,7 @@ static X509 *retrieve_cert(SEXP obj, const char *c_name) {
     return cacrt;
 }
 
-SEXP PKI_verify_cert(SEXP sCA, SEXP sCert) {
+SEXP PKI_verify_cert(SEXP sCA, SEXP sCert, SEXP sDefault, SEXP sPart) {
     X509 *cert;
     X509_STORE *store;
     X509_STORE_CTX *ctx;
@@ -90,15 +92,53 @@ SEXP PKI_verify_cert(SEXP sCA, SEXP sCert) {
     PKI_init();
     cert = retrieve_cert(sCert, "");
     store = X509_STORE_new();
+
+    if (Rf_asInteger(sDefault) > 0)
+	X509_STORE_set_default_paths(store);
+
+    /* highly recommended (and default since OpenSSL 1.1.0) to avoid
+       breakage of chains like the famous Let's Encrypt 2021 sanfu
+       or Sectigo */
+#ifdef X509_V_FLAG_TRUSTED_FIRST
+    X509_STORE_set_flags(store, X509_V_FLAG_TRUSTED_FIRST);
+#endif
+
+#ifdef X509_V_FLAG_PARTIAL_CHAIN
+    if (Rf_asInteger(sPart) > 0)
+	X509_STORE_set_flags(store, X509_V_FLAG_PARTIAL_CHAIN);
+#endif
+
     if (TYPEOF(sCA) == VECSXP) {
 	int i;
 	for (i = 0; i < LENGTH(sCA); i++)
 	    X509_STORE_add_cert(store, retrieve_cert(VECTOR_ELT(sCA, i),"CA "));
-    } else
+    } else if (sCA != R_NilValue)
 	X509_STORE_add_cert(store, retrieve_cert(sCA, "CA "));
+
     ctx = X509_STORE_CTX_new();
     X509_STORE_CTX_init(ctx, store, cert, NULL);
     rv = X509_verify_cert(ctx);
+
+#if 0 /* we could print of even return the chain, this is how ... */
+    {
+	int j;
+
+	STACK_OF(X509) *chain = X509_STORE_CTX_get1_chain(ctx);
+	int num_untrusted = X509_STORE_CTX_get_num_untrusted(ctx);
+	Rprintf("Chain:\n");
+	for (j = 0; j < sk_X509_num(chain); j++) {
+	    X509 *cert = sk_X509_value(chain, j);
+	    X509_NAME *sname = X509_get_subject_name(cert);
+	    char buf[256];
+	    Rprintf("depth=%d: %s", j, X509_NAME_oneline(sname, buf, sizeof(buf) - 1));
+	    if (j < num_untrusted)
+		Rprintf(" (untrusted)");
+	    Rprintf("\n");
+	}
+	sk_X509_pop_free(chain, X509_free);
+    }
+#endif
+
     X509_STORE_CTX_free(ctx);
     X509_STORE_free(store);
     return ScalarLogical((rv == 1) ? TRUE : FALSE);
@@ -191,6 +231,7 @@ static char cipher_name[32];
 static EVP_CIPHER_CTX *get_cipher(SEXP sKey, SEXP sCipher, int enc, int *transient, SEXP sIV) {
     EVP_CIPHER_CTX *ctx;
     PKI_init();
+
     if (inherits(sKey, "symmeric.cipher")) {
 	if (transient) transient[0] = 0;
 	return (EVP_CIPHER_CTX*) R_ExternalPtrAddr(sCipher);
@@ -199,7 +240,7 @@ static EVP_CIPHER_CTX *get_cipher(SEXP sKey, SEXP sCipher, int enc, int *transie
 	Rf_error("invalid key object");
     else {
 	const char *cipher, *c_key, *c_iv = 0;
-	int key_len;
+	size_t key_len;
 	const EVP_CIPHER *type;
 	if (TYPEOF(sCipher) != STRSXP || LENGTH(sCipher) != 1)
 	    Rf_error("non-RSA key and no cipher is specified");
@@ -231,28 +272,44 @@ static EVP_CIPHER_CTX *get_cipher(SEXP sKey, SEXP sCipher, int enc, int *transie
 	else if (!strcmp(cipher, "aes256ofb"))
 	    type = EVP_aes_256_ofb();
 	else if (!strcmp(cipher, "blowfish") || !strcmp(cipher, "bfcbc"))
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	    type = EVP_CIPHER_fetch(PKI_ossl_ctx, "BF-CBC", NULL);
+#else
 	    type = EVP_bf_cbc();
+#endif
 	else if (!strcmp(cipher, "bfecb"))
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	    type = EVP_CIPHER_fetch(PKI_ossl_ctx, "BF-ECB", NULL);
+#else
 	    type = EVP_bf_ecb();
+#endif
 	else if (!strcmp(cipher, "bfofb"))
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	    type = EVP_CIPHER_fetch(PKI_ossl_ctx, "BF-OFB", NULL);
+#else
 	    type = EVP_bf_ofb();
+#endif
 	else if (!strcmp(cipher, "bfcfb"))
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+	    type = EVP_CIPHER_fetch(PKI_ossl_ctx, "BF-CFB", NULL);
+#else
 	    type = EVP_bf_cfb();
+#endif
 	else Rf_error("unknown cipher `%s'", CHAR(STRING_ELT(sCipher, 0)));
 
 	if (TYPEOF(sIV) == STRSXP) {
 	    if (LENGTH(sIV) != 1)
 		Rf_error("invalid IV - if used must be a string (or raw), but is string vector of length %d", (int) LENGTH(sIV));
 	    c_iv = CHAR(STRING_ELT(sIV, 0));
-	    int req_len = EVP_CIPHER_iv_length(type);
-	    int iv_len = strlen(c_iv);
+	    size_t req_len = (size_t) EVP_CIPHER_iv_length(type);
+	    size_t iv_len = strlen(c_iv);
 	    if (iv_len < req_len)
-		Rf_error("insufficient IV - must be %d bytes long", req_len);
+		Rf_error("insufficient IV - must be %u bytes long", (unsigned int) req_len);
 	} else if (TYPEOF(sIV) == RAWSXP) {
 	    c_iv = (const char *) RAW(sIV);
-	    int req_len = EVP_CIPHER_iv_length(type);
-	    if (LENGTH(sIV) < req_len)
-		Rf_error("insufficient IV - must be %d bytes long", req_len);
+	    size_t req_len = (size_t) EVP_CIPHER_iv_length(type);
+	    if (((size_t) LENGTH(sIV)) < req_len)
+		Rf_error("insufficient IV - must be %u bytes long", (unsigned int) req_len);
 	} else if (sIV != R_NilValue)
 	    Rf_error("invalid IV - must be NULL (no/empty IV), a string or a raw vector of sufficient length for the cipher");
 
@@ -261,10 +318,11 @@ static EVP_CIPHER_CTX *get_cipher(SEXP sKey, SEXP sCipher, int enc, int *transie
 	    key_len = strlen(c_key);
 	} else {
 	    c_key = (const char*) RAW(sKey);
-	    key_len = LENGTH(sKey);
+	    key_len = (size_t) LENGTH(sKey);
 	}
-	if (key_len < EVP_CIPHER_key_length(type))
-	    Rf_error("key is too short (%d bytes) for the cipher - need %d bytes", key_len, EVP_CIPHER_key_length(type));
+	if (key_len < (size_t) EVP_CIPHER_key_length(type))
+	    Rf_error("key is too short (%u bytes) for the cipher - need %d bytes",
+		     (unsigned int) key_len, (int) EVP_CIPHER_key_length(type));
 	ctx = EVP_CIPHER_CTX_new();
 	if (!ctx)
 	    Rf_error("cannot allocate memory for cipher");
@@ -277,6 +335,7 @@ static EVP_CIPHER_CTX *get_cipher(SEXP sKey, SEXP sCipher, int enc, int *transie
     }
 }
 
+#if 0
 static void PKI_free_cipher(SEXP sCipher) {
     EVP_CIPHER_CTX *ctx = (EVP_CIPHER_CTX*) R_ExternalPtrAddr(sCipher);
     if (ctx)
@@ -284,6 +343,7 @@ static void PKI_free_cipher(SEXP sCipher) {
 }
 
 /* FIXME: this is exposed as C symbol but not actually used anywhere ... ?!? */
+/* it is not longer registered anyway ... */
 SEXP PKI_sym_cipher(SEXP sKey, SEXP sCipher, SEXP sEncrypt, SEXP sIV) {
     SEXP res;
     int transient_cipher = 0;
@@ -293,10 +353,11 @@ SEXP PKI_sym_cipher(SEXP sKey, SEXP sCipher, SEXP sEncrypt, SEXP sIV) {
 	return sCipher;
     res = PROTECT(R_MakeExternalPtr(ctx, R_NilValue, R_NilValue));
     R_RegisterCFinalizerEx(res, PKI_free_cipher, TRUE);
-    setAttrib(res, install("class"), mkString("symmetric.cipher"));
-    UNPROTECT(1);
-    return res;    
+    setAttrib(res, R_ClassSymbol, PROTECT(mkString("symmetric.cipher")));
+    UNPROTECT(2);
+    return res;
 }
+#endif
 
 SEXP PKI_encrypt(SEXP what, SEXP sKey, SEXP sCipher, SEXP sIV) {
     SEXP res;
@@ -400,13 +461,14 @@ SEXP PKI_decrypt(SEXP what, SEXP sKey, SEXP sCipher, SEXP sIV) {
 SEXP PKI_digest(SEXP sWhat, SEXP sMD) {
     SEXP res;
     unsigned char hash[32]; /* really, at most 20 bytes are needed */
-    int len, md = asInteger(sMD);
+    size_t len, what_len;
+    int md = asInteger(sMD);
     const unsigned char *what;
-    int what_len;
+
     PKI_init();
     if (TYPEOF(sWhat) == RAWSXP) {
 	what = (const unsigned char*) RAW(sWhat);
-	what_len = LENGTH(sWhat);
+	what_len = (size_t) XLENGTH(sWhat);
     } else if (TYPEOF(sWhat) == STRSXP) {
 	if (LENGTH(sWhat) < 1) return allocVector(RAWSXP, 0); /* good? */
 	what = (const unsigned char*) CHAR(STRING_ELT(sWhat, 0));
@@ -419,9 +481,9 @@ SEXP PKI_digest(SEXP sWhat, SEXP sMD) {
 	len = SHA_DIGEST_LENGTH;
 	break;
     case PKI_SHA256:
-  SHA256(what, what_len, hash);
-  len = SHA256_DIGEST_LENGTH;
-  break;
+	SHA256(what, what_len, hash);
+	len = SHA256_DIGEST_LENGTH;
+	break;
     case PKI_MD5:
 	MD5(what, what_len, hash);
 	len = MD5_DIGEST_LENGTH;
@@ -679,17 +741,116 @@ SEXP PKI_get_subject(SEXP sCert) {
     PKI_init();
     cert = retrieve_cert(sCert, "");
     if (X509_NAME_print_ex(mem, X509_get_subject_name(cert), 0, (XN_FLAG_ONELINE | ASN1_STRFLGS_UTF8_CONVERT) & ~ASN1_STRFLGS_ESC_MSB) < 0) {
-      BIO_free(mem);
+	BIO_free(mem);
 	Rf_error("X509_NAME_print_ex failed with %s", ERR_error_string(ERR_get_error(), NULL));
     }
     len = BIO_get_mem_data(mem, &txt);
-    if (len < 0) {
-      BIO_free(mem);
-      Rf_error("cannot get memory buffer, %s", ERR_error_string(ERR_get_error(), NULL));
+    if (len < 0 || len > 2147483646) {
+	BIO_free(mem);
+	Rf_error("cannot get memory buffer, %s", ERR_error_string(ERR_get_error(), NULL));
     }
     res = PROTECT(allocVector(STRSXP, 1));
-    SET_STRING_ELT(res, 0, mkCharLenCE(txt, len, CE_UTF8));
+    SET_STRING_ELT(res, 0, mkCharLenCE(txt, (int) len, CE_UTF8));
     UNPROTECT(1);
     BIO_free(mem);
     return res;
 }
+
+#include <time.h>
+
+static char cibuf[512];
+
+static double ASN1_TIME2d(const ASN1_TIME* time) {
+    int pday, psec;
+    ASN1_TIME *epoch;
+    double d;
+
+#if OPENSSL_VERSION_NUMBER < 0x10002000L
+    Rf_warning("OpenSSL is too old and does not support ASN1 time differences");
+    return NA_REAL;
+#else
+    epoch = ASN1_TIME_set(0, 0);
+    ASN1_TIME_diff(&pday, &psec, epoch, time);
+    ASN1_STRING_free(epoch);
+
+    d = (double) pday;
+    d *= 86400.0;
+    d += (double) psec;
+    return d;
+#endif
+}
+
+SEXP PKI_get_cert_info(SEXP sCert) {
+#define FPLEN 20 /* size of the fingerprint - here SHA1 */
+    const EVP_MD *digest = EVP_sha1();
+    SEXP res = PROTECT(Rf_allocVector(VECSXP, 5));
+    int rc;
+    unsigned len;
+    X509 *cert;
+    double *ts;
+    PKI_init();
+    cert = retrieve_cert(sCert, "");
+    cibuf[sizeof(cibuf) - 1] = 0;
+    *cibuf = 0;
+    X509_NAME_oneline(X509_get_subject_name(cert), cibuf, sizeof(cibuf) - 1);
+    SET_VECTOR_ELT(res, 0, Rf_mkString(cibuf));
+    X509_NAME_oneline(X509_get_issuer_name(cert), cibuf, sizeof(cibuf) - 1);
+    SET_VECTOR_ELT(res, 1, Rf_mkString(cibuf));
+
+    len = FPLEN;
+    rc = X509_digest(cert, digest, (unsigned char*) cibuf, &len);
+    if (rc && len == FPLEN) {
+	SEXP sFP;
+	SET_VECTOR_ELT(res, 2, (sFP = allocVector(RAWSXP, len)));
+	memcpy(RAW(sFP), cibuf, len);
+    }
+
+    ts = REAL(SET_VECTOR_ELT(res, 3, Rf_allocVector(REALSXP, 2)));
+    ts[0] = ASN1_TIME2d(X509_get_notBefore(cert));
+    ts[1] = ASN1_TIME2d(X509_get_notAfter(cert));
+
+    SET_VECTOR_ELT(res, 4, Rf_ScalarLogical(X509_check_ca(cert)));
+    UNPROTECT(1);
+    return res;
+}
+
+/* NOTE: we are intentionally not using macros since thay may not match the
+         actual run-time version. The only exception is OPENSSL_VERSION_TEXT
+	 where we have no other choice */
+SEXP PKI_engine_info(void) {
+    char sver[48];
+    const char *names[] = { "engine", "version", "description", "" };
+    SEXP res = PROTECT(mkNamed(VECSXP, names));
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+    unsigned long ver = OpenSSL_version_num();
+#else
+    unsigned long ver = OPENSSL_VERSION_NUMBER;
+#endif
+#ifdef LIBRESSL_VERSION_NUMBER
+    SET_VECTOR_ELT(res, 0, mkString("libressl"));
+#else
+    SET_VECTOR_ELT(res, 0, mkString("openssl"));
+#endif
+    sver[sizeof(sver) - 1] = 0;
+    snprintf(sver, sizeof(sver) - 1, "%u.%u", (unsigned int) (ver >> 28), (unsigned int) ((ver >> 20) & 255));
+    SET_VECTOR_ELT(res, 1, ScalarReal(atof(sver)));
+#if OPENSSL_VERSION_NUMBER >= 0x30000000L
+    SET_VECTOR_ELT(res, 2, mkString(OpenSSL_version(OPENSSL_FULL_VERSION_STRING)));
+#else
+#   ifdef OPENSSL_VERSION_TEXT
+    SET_VECTOR_ELT(res, 2, mkString(OPENSSL_VERSION_TEXT));
+#   else
+    snprintf(sver, sizeof(sver) - 1, "%s %d.%d.%d%c",
+#      ifdef LIBRESSL_VERSION_NUMBER
+	     "LibreSSL",
+#      else
+	     "OpenSSL",
+#      endif
+	     (unsigned int) (ver >> 28), (unsigned int) ((ver >> 20) & 255),
+	     (unsigned int) ((ver >> 12) & 255), (char) (((ver >> 8) & 31) + 0x60));
+    SET_VECTOR_ELT(res, 2, mkString(sver));
+#   endif
+#endif
+    UNPROTECT(1);
+    return res;
+}
